#' Generate C++ function to calculate V
#' 
#' Returns an R function that calls a C++ function which calculates V efficiently.
#' 
#' V must be a list of functions.
#' 
#' @param apollo_probabilities Likelihood function of the whole model.
#' @param apollo_beta Named numeric vector of parameters to be estimated.
#' @param apollo_inputs List of arguments and settings generated by \link{apollo_validateInputs}.
#' @param V Named list of functions.
#' 
#' @return A function that receives three arguments:
#'         \itemize{
#'           # DataFrame db, NumericVector b, List drw
#'           \item \code{db}: Data.frame containing observations.
#'           \item \code{b}: Named numeric vector containing paramters to be estimated.
#'           \item \code{drw}: Named list containing draws, as contained inside \code{apollo_inputs}.
#'         }
#'         And returns a named list with numeric elements.
#'         
#' @importFrom Rcpp cppFunction
apollo_cppV <- function(apollo_probabilities, apollo_beta, apollo_inputs, V){
  # Check that all elements of V and randCoeff are functions
  if(!all(sapply(V, is.function))) stop("All utilities should be functions for them to be converted to C++")
  if(!all(sapply(apollo_inputs$randCoeff, is.function))) stop("All definitions of random coefficients should be functions for them to be converted to C++")
  
  # Build variable definitions
  vars <- varList(apollo_probabilities, apollo_beta, apollo_inputs, V)
  nObs <- nrow(apollo_inputs$database)
  nDrw <- apollo_inputs$apollo_draws$interNDraws
  dec <- c()
  for(b in vars$b) dec <- c(dec, paste0('  double ', b, ' = b["', b, '"];') )
  for(x in vars$x) dec <- c(dec, paste0('  const NumericVector ', x, ' = db["', x, '"];') )
  for(d in vars$d) dec <- c(dec, paste0('  const NumericMatrix ', d, ' = drw["', d, '"];') )
  for(r in vars$r[,1]) dec <- c(dec, paste0('  NumericVector ', r, '(', nObs, ');') )
  for(p in vars$p[,1]) dec <- c(dec, paste0('  NumericVector ', p, '(', nObs, ');') )
  for(i in names(V)) dec <- c(dec, paste0('  NumericMatrix V_', i, '(', nObs, ',', nDrw, ');') )
  dec <- paste0(dec, collapse="\n")
  
  # Replace x^y -> pow(x,y)
  # Not implemented
  
  # Build loop
  loop <- paste0("  for(int d=0; d<", nDrw, "; ++d){")
  loop <- c(loop, paste0("   ", vars$r[,1], " = ", vars$r[,2], ";") )
  loop <- c(loop, paste0("   ", vars$p[,1], " = ", vars$p[,2], ";") )
  loop <- c(loop, paste0("   V_", vars$v[,1], "(_,d) = ", vars$v[,2], ";") )
  for(l in 2:length(loop)) for(j in vars$d) if(regexpr(j, loop[l])>=0) loop[l] <- gsub(j, paste0(j,"(_,d)"), loop[l])
  loop <- c(loop, "  }")
  loop <- paste0(loop, collapse="\n")
  
  # Put utilities in a list
  # List L = List::create(Named("name1") = v1 , Named("name2) = v2);
  vlst <- "  List V = List::create("
  for(i in names(V)[1:(length(V)-1)]) vlst <- c(vlst, 'Named("', i, '") = V_', i, ", ")
  i <- names(V)[length(V)]
  vlst <- c(vlst, 'Named("', i, '") = V_', i, ");")
  vlst <- paste0(vlst, collapse="")
  
  # Put everything together
  header <- paste(c("#include <Rcpp.h>", 
                    "using namespace Rcpp;",
                    "// [[Rcpp::export]]",
                    "List utilPre(DataFrame db, NumericVector b, List drw) {"),
                  sep="", collapse="\n")
  script <- paste0(c(header, dec, loop, vlst, "  return V; \n}"), collapse="\n\n")
  
  # Generate C++ code and return
  cppF <- tryCatch(Rcpp::cppFunction(script), error=function(e) return(NULL))
  return(cppF)
}