#' Calculates the probability of an ordered logit model
#'
#' Calculates the probabilities of an ordered logit model and can also perform other operations based on the value of the \code{functionality} argument.
#'
#' This function estimates an ordered logit model of the type:
#' y* = V + epsilon
#' outcomeOrdered =  1 if   -Inf < y* < tau[1]
#'      2 if tau[1] < y* < tau[2]
#'      ...
#'      maxLvl if tau[length(tau)] < y* < +Inf
#' Where epsilon is distributed standard logistic, and the values 1, 2, ..., maxLvl can be
#' replaces by coding[1], coding[2], ..., coding[maxLvl].
#' The behaviour of the function changes depending on the value of the \code{functionality} argument.
#' @param ol_settings List of settings for the OL model. It should include the following.
#'                   \itemize{
#'                     \item \code{outcomeOrdered} Numeric vector. Dependant variable. The coding of this variable is assumed to be from 1 to the maximum number of different levels. For example, if the ordered response has three possible values: "never", "sometimes" and "always", then it is assumed that outcomeOrdered contains "1" for "never", "2" for "sometimes", and 3 for "always". If another coding is used, then it should be specified using the \code{coding} argument.
#'                     \item \code{V} Numeric vector. A single explanatory variable (usually a latent variable). Must have the same number of rows as outcomeOrdered.
#'                     \item \code{tau} List of numeric vector/matrix/3-dim arrays. Thresholds. As many elements as number of different levels in the dependent variable - 1. Extreme thresholds are fixed at -inf and +inf. Mixing is allowed in thresholds.
#'                     \item \code{coding} Numeric or character vector. Optional argument. Defines the order of the levels in \code{outcomeOrdered}. The first value is associated with the lowest level of \code{outcomeOrdered}, and the last one with the highest value. If not provided, is assumed to be \code{1:(length(tau) + 1)}.
#'                     \item \code{rows} Boolean vector. TRUE if a row must be considered in the calculations, FALSE if it must be excluded. It must have length equal to the length of argument \code{outcomeOrdered}. Default value is \code{"all"}, meaning all rows are considered in the calculation.
#'                     \item \code{componentName} Character. Name given to model component.
#'                   }
#' @param functionality Character. Can take different values depending on desired output.
#'                      \itemize{
#'                        \item \code{"estimate"} Used for model estimation.
#'                        \item \code{"prediction"} Used for model predictions.
#'                        \item \code{"validate"} Used for validating input.
#'                        \item \code{"zero_LL"} Used for calculating null likelihood.
#'                        \item \code{"conditionals"} Used for calculating conditionals.
#'                        \item \code{"output"} Used for preparing output after model estimation.
#'                        \item \code{"raw"} Used for debugging.
#'                      }
#' @return The returned object depends on the value of argument \code{functionality} as follows.
#'         \itemize{
#'           \item \strong{\code{"estimate"}}: vector/matrix/array. Returns the probabilities for the chosen alternative for each observation.
#'           \item \strong{\code{"prediction"}}: List of vectors/matrices/arrays. Returns a list with the probabilities for all possible levels, with an extra element for the probability of the chosen alternative.
#'           \item \strong{\code{"validate"}}: Same as \code{"estimate"}, but it also runs a set of tests to validate the function inputs.
#'           \item \strong{\code{"zero_LL"}}: Not implemented. Returns a vector of NA with as many elements as observations.
#'           \item \strong{\code{"conditionals"}}: Same as \code{"estimate"}
#'           \item \strong{\code{"output"}}: Same as \code{"estimate"} but also writes summary of input data to internal Apollo log.
#'           \item \strong{\code{"raw"}}: Same as \code{"prediction"}
#'         }
#' @export
apollo_ol  <- function(ol_settings, functionality){
  if(is.null(ol_settings[["componentName"]])) ol_settings[["componentName"]]="OL"
  componentName=ol_settings[["componentName"]]
  
  if(is.null(ol_settings[["outcomeOrdered"]])) stop("The ol_settings list for model component \"",componentName,"\" needs to include an object called \"outcomeOrdered\"!")
  if(is.null(ol_settings[["V"]])             ) stop("The ol_settings list for model component \"",componentName,"\" needs to include an object called \"V\"!")
  if(is.null(ol_settings[["tau"]])           ) stop("The ol_settings list for model component \"",componentName,"\" needs to include an object called \"tau\"!")
  if(is.null(ol_settings[["coding"]])        ) ol_settings[["coding"]] = NULL
  if(is.null(ol_settings[["rows"]])          ) ol_settings[["rows"]]   = "all"

  outcomeOrdered=ol_settings[["outcomeOrdered"]]
  V      = ol_settings[["V"]]
  tau    = ol_settings[["tau"]]
  coding = ol_settings[["coding"]]
  rows   = ol_settings[["rows"]]
  nObs <- tryCatch(nrow( get("apollo_inputs", parent.frame(), inherits=FALSE)$database ),
                   error=function(e){
                     lenV <- sapply(V, function(v) ifelse(is.array(v), dim(v)[1], length(v)) )
                     lenC <- length(outcomeOrdered)
                     return(max(lenV, lenC))
                   })
  
  ### NEW LINES 27/3 
  ### Format of tau
  tau_original=tau
  tau=list()
  if(!(is.vector(tau_original)||is.list(tau_original)||is.matrix(tau_original))) stop("Thresholds for ordered logit for model component \"",componentName,"\" need to be a vector, matrix, or list!")
  if(is.vector(tau_original)){
    if(length(tau_original)==1) stop("If provided as scalars, need at least two thresholds for ordered logit for model component \"",componentName,"\"!")
    tau=as.list(tau_original)
    }
  if(is.matrix(tau_original)){
    if(nrow(tau_original)!=nObs) stop("If provided as a matrix, need one value per observation in the data for each threshold for ordered logit for model component \"",componentName,"\"!")
    tau=lapply(seq_len(ncol(tau_original)), function(i) tau_original[,i])
    }
  if(is.list(tau_original)){
    if(length(tau_original)==1) stop("If provided as a list, need at least two elements in list of thresholds for ordered logit for model component \"",componentName,"\"!")
    mixing <- tryCatch(get("apollo_inputs", parent.frame(), inherits=TRUE)$apollo_control$mixing,
                               error = function(e) return(NA))
    apollo_draws <- tryCatch(get("apollo_inputs", parent.frame(), inherits=TRUE)$apollo_draws,
                       error = function(e) return(NA))
    if(any(sapply(tau_original,is.list))) stop("If provided as a list, elements in list of thresholds for ordered logit for model component \"",componentName,"\" cannot be lists themselves!")
    if(!mixing && (any(is.matrix(tau_original)) || any(is.array(tau_original)))) stop("Model uses no mixing, so thresholds for ordered logit for model component \"",componentName,"\" should not be given as matrices or arrays!")
    max_dim=max(sapply((sapply(tau_original,dim)),length))
    if(mixing){
      interNDraws=apollo_draws$interNDraws
      intraNDraws=apollo_draws$intraNDraws
    }
    if(max_dim==3 && intraNDraws==0) stop("Some thresholds for ordered logit for model component \"",componentName,"\" are given as arrays despite the model not using intra-individual mixing!")
    k=1
    while(k<=length(tau_original)){
      if(is.vector(tau_original[[k]])){
        tau[[k]]=tau_original[[k]]
        if(!(length(tau[[k]])%in%c(1,nObs))) stop("Threshold ",k," for ordered logit for model component \"",componentName,"\" is given as a vector and should have one entry per observation in the data!")
        if(length(tau[[k]])==1) tau[[k]]=rep(tau[[k]],nObs)  
        if(max_dim==2) tau[[k]]=matrix(tau[[k]],nrow=nObs,ncol=interNDraws,byrow=FALSE)
        if(max_dim==3) tau[[k]]=array(tau[[k]],dim=c(nObs,interNDraws,intraNDraws))
      }
      if(is.array(tau_original[[k]])){
        tau[[k]]=tau_original[[k]]
        if(!(nrow(tau[[k]])==nObs)) stop("Threshold ",k," for ordered logit for model component \"",componentName,"\" is given as a matrix/array and should have one entry per observation in the data!")
        if(dim(tau[[k]])[2]!=interNDraws) stop("Threshold ",k," for ordered logit for model component \"",componentName,"\" has a number of columns that is not equal to the number of inter-individual draws in the model!")
        if(length(dim(tau[[k]]))!=3) tau[[k]]=array(tau[[k]],dim=c(nObs,interNDraws,intraNDraws))
        if(dim(tau[[k]])[3]!=intraNDraws) stop("Threshold ",k," for ordered logit for model component \"",componentName,"\" has a number of columns in the third dimension that is not equal to the number of intra-individual draws in the model!")
      }
      k=k+1
    }
  }
  #tau <- lapply(tau, apollo_keepRows, r=rows) # 13 May 2020
  
  ### moved here from later on
  if(is.null(coding)){
   coding <- 1:(length(tau)+1)  
   if(functionality=="validate") cat("\nNo coding provided for ordered logit for model component \"",componentName,"\",\n so assuming outcomeOrdered goes from 1 to ",max(coding),"\n",sep="")
  }
  ### FINISHED NEW LINES HERE
  
  ### Format checks
  # outcomeOrdered
  test <- is.vector(outcomeOrdered) && (length(outcomeOrdered)==nObs || length(outcomeOrdered)==1)
  if(!test) stop("The \"outcomeOrdered\" argument for model component \"",componentName,"\" needs to be a scalar or a vector with one entry per observation in the \"database\"")
  # V
  test <- is.numeric(V) && (is.vector(V) | is.array(V))
  lenV <- ifelse(is.array(V), dim(V)[1], length(V))
  test <- test && (lenV==nObs | lenV==1)
  if(!test) stop("\"V\" for model component \"",componentName,"\" needs to be a scalar or a vector/matrix/cube with one row per observation in the \"database\"")  
  # rows
  test <- is.vector(rows) && ( (is.logical(rows) && length(rows)==nObs) || (length(rows)==1 && rows=="all") )
  if(!test) stop("The \"rows\" argument for model component \"",componentName,"\" needs to be \"all\" or a vector of boolean statements with one entry per observation in the \"database\"")
  # functionality
  test <- functionality %in% c("estimate","prediction","validate","zero_LL","conditionals","output","raw")
  if(!test) stop("Non-permissable setting for \"functionality\" for model component \"",componentName,"\"")
  # coding
  test <- is.null(coding) || (is.vector(coding))
  if(!test) stop("Argument 'coding', if provided for model component \"",componentName,"\", must be a vector.")
  
  ### Get number of observations
  nObs <- tryCatch(nrow( get("apollo_inputs", parent.frame(), inherits=TRUE)$database ),
                   error=function(e){
                     lenV <- sapply(V, function(v) ifelse(is.array(v), dim(v)[1], length(v)) )
                     lenC <- length(outcomeOrdered)
                     return(max(lenV, lenC))
                   })
  
  ### Validate input
  #if(!is.vector(outcomeOrdered)) stop("Argument 'outcomeOrdered' must be a vector of length nObs.")
  #if(!is.numeric(V) || !(is.array(V) | is.vector(V))) stop("Argument 'V' must be a numeric vector or array.")
  #if(!is.null(coding) && (!is.vector(coding) | is.null(names(coding)))) stop("Argument 'coding', if provided, must be a named vector.")
  ### CORRECTED NEXT LINE
  ### if(rows!="all" & ( length(rows)!=nObs | !is.logical(rows) )) stop("Argument 'rows', if provided, must be \"all\" or a vector of boolean statements of length nObs.")
  #if(!( length(rows)==1 && (rows=="all")) & ( length(rows)!=nObs | !is.logical(rows) )) stop("Argument 'rows', if provided, must be \"all\" or a vector of boolean statements of length nObs.")
  
  ### Filter rows
  if(length(rows)==1 && rows=="all") rows <- rep(TRUE, length(nObs))
  if(any(!rows)){
    outcomeOrdered <- outcomeOrdered[rows]
    V <- apollo_keepRows(V, rows)
    tau <- lapply(tau, apollo_keepRows, r=rows) # 13 May 2020
  }
  nObs <- max(ifelse(is.array(V), dim(V)[1], length(V)) , length(outcomeOrdered))
  

  # ############################## #
  #### functionality="validate" ####
  # ############################## #

  if(functionality=="validate"){
    apollo_control <- tryCatch(get("apollo_inputs", parent.frame(), inherits=FALSE)$apollo_control,
                               error = function(e) list(noValidation=FALSE, noDiagnostics=FALSE))
    
    if(!apollo_control$noValidation){
      ### REMOVED 27/3 as now earlier and different
      ##if(!is.vector(tau)) stop("Thresholds for ordered logit for model component \"",componentName,"\" need to be a vector (no random components allowed)!")
      values_present=unique(outcomeOrdered)
      ### REMOVED 27/3 as now earlier
    ###if(is.null(coding)){
    ###    coding <- 1:(length(tau)+1)  
    ###    cat("\nNo coding provided for ordered logit for model component \"",componentName,"\",\n so assuming outcomeOrdered goes from 1 to",max(coding),"\n")
    ###  }
      if(!(all(values_present %in% coding ))) stop("Some levels in 'outcomeOrdered' do not exist in 'coding' for model component \"",componentName,"\" !")
      if(!(all(coding %in% values_present ))) stop("Some levels in 'coding' do not exist in 'outcomeOrdered' for model component \"",componentName,"\"!")
      if( (length(tau)+1)!=length(coding) ) stop("Threshold vector length +1 does not match number of elements in argument 'coding' for model component \"",componentName,"\".")
    }
    
    if(!apollo_control$noDiagnostics){
      choicematrix <- t(as.matrix(table(outcomeOrdered)))
      choicematrix <- rbind(choicematrix, choicematrix[1,]/nObs*100)
      rownames(choicematrix) <- c("Times chosen", "Percentage chosen overall")
      apolloLog <- tryCatch(get("apollo_inputs", parent.frame(), inherits=TRUE )$apolloLog, error=function(e) return(NA))
      apollo_addLog(title   = paste0("Overview of choices for model component \"",componentName,"\""), 
                    content = round(choicematrix,2), apolloLog)
    }

    testL = apollo_ol(ol_settings, functionality="estimate")
    if(all(testL==0)) stop("\nAll observations have zero probability at starting value for model component \"",componentName,"\"")
    if(any(testL==0)) cat("\nSome observations have zero probability at starting value for model component \"",componentName,"\"")
    return(invisible(testL))
  }

  # ########################################################## #
  #### functionality="zero_LL"                              ####
  # ########################################################## #

  if(functionality=="zero_LL"){
    P <- rep(NA, nObs)
    if(any(!rows)) P <- apollo_insertRows(P, rows, 1)
    return(P)
  }

  # ########################################################## #
  #### functionality="estimate | conditionals | raw"        ####
  # ########################################################## #

  if(functionality %in% c("estimate","conditionals")){

    ###if(is.null(coding)) coding <- 1:(length(tau)+1)
    map <- stats::setNames(1:length(coding), coding)
    outcomeOrdered2 <- map[as.character(outcomeOrdered)]

    tau <- c(-Inf,tau,Inf)

    ### NEW LINES 27/3
    ### extract appropriate taus
    #tau1 <- Reduce('+',lapply(as.list(2:(length(coding))),FUN=function(i) (outcomeOrdered2+1==i)*tau[[i]]))
    #tau1[outcomeOrdered2==5]=Inf
    #tau2 <- Reduce('+',lapply(as.list(2:length(coding)),FUN=function(i) (outcomeOrdered2==i)*tau[[i]]))
    #tau2[outcomeOrdered2==1]=-Inf
    
    
    ### CORRECTED 24/4/20
    tau1 <- Reduce('+',lapply(as.list(2:(length(coding))),FUN=function(i) (outcomeOrdered2+1==i)*tau[[i]]))
    tau1[outcomeOrdered2==(length(coding))]=Inf
    tau2 <- Reduce('+',lapply(as.list(2:length(coding)),FUN=function(i) (outcomeOrdered2==i)*tau[[i]]))
    tau2[outcomeOrdered2==1]=-Inf
    
    p <- 1/(1 + exp(V-tau1)) - 1/(1 + exp(V-tau2))
    
    ###OLD
    ###p <- 1/(1 + exp(V-tau[outcomeOrdered2+1])) - 1/(1 + exp(V-tau[outcomeOrdered2]))
    ### FINISHED NEW LINES
    
    if(any(!rows)) p <- apollo_insertRows(p, rows, 1) # insert excluded rows with value 1
    
    return(p)
  }

  # ########################################################## #
  #### functionality="prediction"                           ####
  # ########################################################## #
  if(functionality %in% c("prediction", "raw")){
    
    ###if(is.null(coding)) coding <- 1:(length(tau)+1)
    tau <- c(-Inf,tau,Inf)
    
    p = list()
    ### NEW LINES 27/3
    for(j in 1:(length(tau)-1)) p[[j]] = 1/(1 + exp(V-tau[[j+1]])) - 1/(1 + exp(V-tau[[j]]))
    ### OLD
    ### for(j in 1:(length(tau)-1)) p[[j]] = 1/(1 + exp(V-tau[j+1])) - 1/(1 + exp(V-tau[j]))
    
    if(is.null(coding)) coding <- 1:(length(tau)-1)
    names(p) <- coding
    
    # insert excluded rows with value NA
	  if(any(!rows)) p <- lapply(p, apollo_insertRows, r=rows, val=NA)
    
    # Add chosen alternative (unless outcomeOrdered is NA)
    if(!(length(outcomeOrdered)==1 && is.na(outcomeOrdered))){
      p[["chosen"]] <- apollo_ol(ol_settings, functionality="estimate")
    }

    return(p)
  }

  # ############################# #
  #### functionality="output"  ####
  # ############################# #
  if(functionality=="output"){

    p <- apollo_ol(ol_settings, functionality="estimate")
    
    choicematrix <- t(as.matrix(table(outcomeOrdered)))
    choicematrix <- rbind(choicematrix, choicematrix[1,]/nObs*100)
    rownames(choicematrix) <- c("Times chosen", "Percentage chosen overall")
    
    apolloLog <- tryCatch(get("apollo_inputs", parent.frame(), inherits=TRUE )$apolloLog, error=function(e) return(NA))
    apollo_addLog(title   = paste0("Overview of choices for model component \"",componentName,"\""), 
                  content = round(choicematrix,2), apolloLog)
    apollo_reportModelTypeLog(modelType="OL", apolloLog)
    
    return(p)
  }


}
