#' Lists variables names and definitions used in V
#' 
#' Returns a list containing the names and definitions of variables used in V
#' 
#' It looks for variables in apollo_beta, apollo_randCoeff, draws, and 
#' apollo_probabilities. It returns them in a list ordered by origin.
#' 
#' @param apollo_probabilities Likelihood function of the whole model.
#' @param apollo_beta Named numeric vector of parameters to be estimated.
#' @param apollo_inputs List of arguments and settings generated by \link{apollo_validateInputs}.
#' @param V Named list of functions.
#' 
#' @return A list containing the following elements (all of type character):
#'         \itemize{
#'           \item \code{b}: Vector with the names of variables contained in \code{apollo_beta}.
#'           \item \code{x}: Vector with the names of variables contained in \code{database}.
#'           \item \code{d}: Vector with the names of variables contained in \code{draws}.
#'           \item \code{r}: Matrix with the names (first column) and definitions (second column) of variables contained in \code{apollo_randCoeff}.
#'           \item \code{p}: Matrix with the names (first column) and definitions (second column) of variables contained in \code{apollo_probabilities}.
#'           \item \code{v}: Matrix with the names (first column) and definitions (second column) of utilities contained in \code{V}.
#'         }
apollo_varList <- function(apollo_probabilities, apollo_beta, apollo_inputs, V){
  
  # Check there are no assignments inside the V and randCoeff functions
  for(v in V){
    bF <- body(v)
    for(i in 1:length(bF)){
      test <- length(bF[[i]])==3 && (bF[[i]][[1]]=="=" || bF[[i]][[1]]=="<-")
      if(test) stop("Utilities should not contain assignments")
    }
  }
  for(r in rndCoeff){
    bF <- body(r)
    for(i in 1:length(bF)){
      test <- length(bF[[i]])==3 && (bF[[i]][[1]]=="=" || bF[[i]][[1]]=="<-")
      if(test) stop("Random coefficients should not contain assignments")
    }
  }
  
  # Read names and definitions of V
  v <- matrix("", nrow=length(V), ncol=2)
  for(i in 1:length(V)){
    v[i,1] <- names(V)[i]
    v[i,2] <- capture.output(print( body(V[[i]]) ))
  }
  
  # Read explanatory variables names and remove unused ones
  vars.x <- names(apollo_inputs$database)
  drop   <- c()
  for(i in 1:length(vars.x)) if(all(regexpr(vars.x[i], v[,2])<0)) drop <- c(i, drop)
  vars.x <- vars.x[-drop]
  
  # Read parameters names and remove unused ones
  vars.b <- names(apollo_beta)
  drop   <- c()
  for(i in 1:length(vars.b)) if(all(regexpr(vars.b[i], v[,2])<0)) drop <- c(i, drop)
  vars.b <- vars.b[-drop]
  
  # Read draws names and remove unused ones
  if(is.list(apollo_inputs$draws)){
    vars.d <- names(apollo_inputs$draws)
    drop   <- c()
    for(i in 1:length(vars.d)) if(all(regexpr(vars.d[i], v[,2])<0)) drop <- c(i, drop)
    vars.d <- vars.d[-drop]
  }  else vars.d <- c()
  
  # Read random parameters names and remove unused ones
  if(is.function(apollo_inputs$apollo_randCoeff)){
    rndCoeff <- apollo_inputs$apollo_randCoeff
    environment(rndCoeff) <- list2env(c(as.list(apollo_beta), 
                                        apollo_inputs$database, 
                                        apollo_inputs$draws), hash=TRUE)
    rndCoeff <- rndCoeff(apollo_beta, apollo_inputs)
  } else rndCoeff <- NA
  r <- NULL
  if(is.list(rndCoeff)){
    r <- matrix("", nrow=length(rndCoeff), ncol=2)
    r[,1] <- names(rndCoeff)
    for(i in vars.r) r[i,2] <- capture.output(print( body(rndCoeff[[i]]) ) )
    drop   <- c()
    for(i in 1:nrow(r)) if(all(regexpr(r[i,1], v[,2])<0)) drop <- c(i, drop)
    r <- r[-drop,]
  }
  
  # Read names from variables defined in apollo_probabilities and remove unused ones
  bL     <- body(apollo_probabilities)
  vars.p <- c(); def.p <- c()
  for(i in 1:length(bL)){
    # An assignment has three elements: 
    # [1] "=" or "<-", [2] names of new var, [3] value assigned
    bLi  <- bL[[i]]
    test <- length(bLi)==3 && (bLi[[1]]=="=" || bLi[[1]]=="<-")
    test <- test && length(bLi[[2]])==1
    test <- test && !(length(bLi[[3]])>=1 && bLi[[3]][[1]]=="list")
    test <- test && !(length(bLi[[3]])>=1 && bLi[[3]][[1]]=="apollo_avgIntraDraws")
    test <- test && !(length(bLi[[3]])>=1 && bLi[[3]][[1]]=="apollo_panelProd")
    test <- test && !(length(bLi[[3]])>=1 && bLi[[3]][[1]]=="apollo_avgInterDraws")
    test <- test && !(length(bLi[[3]])>=1 && bLi[[3]][[1]]=="apollo_prepareProb")
    if(test){
      vars.p <- c(vars.p, as.character(bL[[i]][[2]]))
      def.p  <- c(def.p, capture.output(print(bL[[i]][[3]])) )
      drop   <- c()
      for(i in 1:length(vars.p)) if(all(regexpr(vars.p[i], v[,2])<0)) drop <- c(i, drop)
      vars.p <- vars.p[-drop]
      def.p  <- def.p[-drop]
    }
    if(length(vars.p)>0) p <- cbind(vars.p, def.p) else p <- null
  }
  
  # Stitch together
  vars <- list(b=vars.b, x=vars.x, d=vars.d, r=r, p=p, v=v)
  return(vars)
}